/*
 *  import_v4l.c
 *
 *  Copyright (C) Thomas streich - February 2002
 *
 *  This file is part of transcode, a video stream processing tool
 *      
 *  transcode is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  transcode is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

//#include "transcode.h"
#include "vcr.h"

#include <stdint.h>

#define MOD_NAME    "import_v4l.so"
#define MOD_VERSION "v0.0.5 (2003-06-11)"
#define MOD_CODEC   "(video) v4l | (audio) PCM"

#define TC_QUIET 0
#define TC_CAP_RGB 1
#define TC_CAP_YUV 2
#define TC_CAP_PCM 3
#define TC_VIDEO 4
#define TC_AUDIO 5
#define TC_STATS 6
#define CODEC_RGB 7
#define CODEC_YUV 8
#define TC_IMPORT_ERROR 9

static int verbose_flag=TC_QUIET;
static int capability_flag=TC_CAP_RGB|TC_CAP_YUV|TC_CAP_PCM;

#define MOD_PRE v4l
//#include "import_def.h"

#define MAX_DISPLAY_PTS 25

//static char *default_audio="/dev/dsp";
//static char *default_video="/dev/video";

static uint32_t aframe_cnt=0;
static uint32_t vframe_cnt=0;

static double aframe_pts0=0, aframe_pts=0;
static double vframe_pts0=0, vframe_pts=0;

static int audio_drop_frames=25;
static int video_drop_frames=0;
static int do_audio = 1;

static int do_resync = 1;

static int verbose = 0;

#include "avilib.h"

typedef struct _transfer_t {

	int   flag;
	FILE   *fd;
  
	int   size;

	char *buffer;
	char *buffer2;

	int attributes;

} transfer_t;

typedef struct _vob_t {
  
  // import info

	char *vmod_probed;
	char *amod_probed;
	char *vmod_probed_xml;       // type of content of xml file
	char *amod_probed_xml;       // type of content of xml file

  
	int verbose;
  
	char *video_in_file;       // video source
	char *audio_in_file;       // audio source
  
	char *nav_seek_file;        // seek/index information

	int in_flag;

	int has_audio;           // stream has probably no audio 
	int has_audio_track;     // stream has probably not requested track
	int has_video;           // stream has probably no video
 
	int lang_code;            // track language
  
	int a_track;
	int v_track;
	int s_track;              // subtitle track id
  
	int sync;                 // audio frame offset for synchronization
	int sync_ms;              // fine tuning audio/video offset

	int dvd_title;
	int dvd_chapter1;
	int dvd_chapter2;
	int dvd_max_chapters;
	int dvd_angle;
 
	int ps_unit;
	int ps_seq1;
	int ps_seq2;

	int ts_pid1;
	int ts_pid2;

	int vob_offset;
	int vob_chunk;
	int vob_chunk_num1;
	int vob_chunk_num2;
	int vob_chunk_max;
	int vob_percentage;

	int vob_psu_num1;
	int vob_psu_num2;

	char *vob_info_file;

	double pts_start;

	double psu_offset;         // psu offset to pass to extsub

	int demuxer;

	long format_flag;          //NTSC=1,PAL=0
	long codec_flag;           //MPEG version ...

	int quality;

	int af6_mode;
  
  // audio stream parameter

	int a_stream_bitrate;     //source stream bitrate
  
	int a_chan;
	int a_bits;
	int a_rate;

	int a_padrate;            // zero padding rate

	int im_a_size;            // import total bytes per audio frame 
	int ex_a_size;            // export total bytes per audio frame 

	int im_a_codec;           // true frame buffer audio codec
	int fixme_a_codec;        // true frame buffer audio codec
  
	int a_leap_frame;
	int a_leap_bytes;

	int a_vbr;                // lame vbr switch

	int a52_mode;

	int dm_bits;
	int dm_chan;

  // video stream parameter

	int v_stream_bitrate;     //source stream bitrate
  
	double fps;               // defaults to PAL: 25 fps
	int im_frc;               // import frame rate code
	double ex_fps;            // defaults to PAL: 25 fps
	int ex_frc;               // export frame rate code
	int hard_fps_flag;        // if this is set, disable smooth drop in demuxer

	int pulldown;             // set 3:2 pulldown flags on MPEG export

	int im_v_height;          // import picture height
	int im_v_width;           // import picture width
	int im_v_size;            // total number of bytes per frame

	int v_bpp;                // defaults to BPP
  
	int im_asr;               // import aspect ratio code  
	int im_par;               // import pixel aspect (code)
	int im_par_width;         // import pixel aspect width
	int im_par_height;        // import pixel aspect height
	int ex_asr;               // export aspect ratio code  
	int ex_par;               // export pixel aspect (code)
	int ex_par_width;         // export pixel aspect width
	int ex_par_height;        // export pixel aspect height
	int attributes;           // more video frame attributes 

	int im_v_codec;           // true frame buffer video codec

	int encode_fields;        // flag

	int dv_yuy2_mode;            // DV decoding option (0.6.5, ThOe)

  // audio frame manipulation info

	int core_a_format;      // internal audio data format

	double volume;          // audio amplitude rescale parameter
	double ac3_gain[3];     // audio amplitude rescale parameter for ac3
	int clip_count;         // total number of bytes clipped after 
                          // volume adjustment 

  // video frame manipulation info

	int core_v_format;      // internal video data format  

	int ex_v_width;         // export picture width
	int ex_v_height;        // export picture height
	int ex_v_size;          // total number of bytes per frame
  
	int reduce_h;             // reductionfactor for frame size 2^reduce
	int reduce_w;             // reductionfactor for frame size 2^reduce

	int resize1_mult;       // row resp. colums multiplier
	int vert_resize1;        // resize height: h -> h - vert_resize*resize1_mult rows
	int hori_resize1;        // resize width:  w -> w - hori_resize*resize1_mult coloums

  
	int resize2_mult;        // row resp. colums multiplier
	int vert_resize2;        // resize height: h -> h + vert_resize*resize2_mult rows
	int hori_resize2;        // resize width:  w -> w + hori_resize*resize2_mult coloums

	int zoom_width;         // zoom width
	int zoom_height;        // zoom height

	double (*zoom_filter)(double);
	double zoom_support;

	int antialias;          // not yet implemented
	int deinterlace;        // not yet implemented
	int decolor;

	double aa_weight;       // anti-alias center pixel weight
	double aa_bias;         // anti-alias horizontal/vertical bias
   
	double gamma;

	int ex_clip_top;
	int ex_clip_bottom;
	int ex_clip_left;
	int ex_clip_right;

	int im_clip_top;
	int im_clip_bottom;
	int im_clip_left;
	int im_clip_right;

	int post_ex_clip_top;
	int post_ex_clip_bottom;
	int post_ex_clip_left;
	int post_ex_clip_right;

	int pre_im_clip_top;
	int pre_im_clip_bottom;
	int pre_im_clip_left;
	int pre_im_clip_right;

  // export info

	char *video_out_file;
	char *audio_out_file;

	avi_t *avifile_in;
	avi_t *avifile_out;
	int avi_comment_fd; // text file to read avi header comments from

	int out_flag;

  // encoding parameter
  
	int divxbitrate;
	int divxkeyframes;
	int divxquality;
	int divxcrispness;
	int divxmultipass;
	int video_max_bitrate;
	char *divxlogfile;

	int min_quantizer;
	int max_quantizer;

	int rc_period;
	int rc_reaction_period;
	int rc_reaction_ratio;

	int divx5_vbv_prof; // profile
  // Video Bitrate Verifier contraints (override profile)
	int divx5_vbv_bitrate;
	int divx5_vbv_size;
	int divx5_vbv_occupancy;

	int mp3bitrate;
	int mp3frequency;
	float mp3quality;         //0=best (very slow).  9=worst (default=5)
	int mp3mode;              //0=joint-stereo, 1=full-stereo, 2=mono

	int bitreservoir;
	char *lame_preset;

	char *audiologfile;
  
	int ex_a_codec;         // audio codec for export module
	int ex_v_codec;         // video codec fcc for export module
  
	char *ex_v_fcc;        //video fourcc string
	char *ex_a_fcc;        //audio fourcc string/identifier
	char *ex_profile_name; //user profile name
  
	int pass_flag;
	int lame_flush;
  
	char *mod_path;

	struct fc_time *ttime; // for framecode parsing, pointer to first struct
	int  ttime_current;    // current time element, needed?

	unsigned int frame_interval; //select every `frame_interval` frames only

	int   chanid;
	char *station_id;

	char *im_v_string;  // pass parameters to import video modules
	char *im_a_string;  // pass parameters to import audio modules
	char *ex_v_string;  // pass parameters to export video modules
	char *ex_a_string;  // pass parameters to export audio modules

	int accel;

	int video_frames_delay; // delay audio by N frames, (video is late)
	float m2v_requant;    // Requantize Factor for mpeg2 video streams

	//mpeg_profile_t mpeg_profile;
  
} vob_t;

// video_grab_init, video_grab_frame, video_grab_close
// v4l_counter_init, v4l_counter_print
// audio_grab_init, audio_grab_frame, audio_grab_close

/* ------------------------------------------------------------ 
 *
 * open stream
 *
 * ------------------------------------------------------------*/

static int v4l_open(transfer_t *param, vob_t *vob)
{
    int fmt = VIDEO_PALETTE_YUV420P;
  
  
  if(param->flag == TC_VIDEO) {
    
    // print out
    if(verbose_flag) printf("[%s] video4linux video grabbing\n", MOD_NAME);
    
    param->fd = NULL;
    
    //set channel_id with vob->v_track 
    //set channel with vob->station or vob->station_id 
    //set device with vob->video_in_file

    //printf("vob->amod_probed (%s)\n", vob->amod_probed);

    /* This check is bogus since amod->probed does not contain what the user
     * specified with -x 

    if ((vob->amod_probed && strlen(vob->amod_probed)>=4 && !strncmp(vob->amod_probed, "null", 4)) 
	    || !vob->amod_probed) {
	printf("NO AUDIO\n");
	do_audio = 0;
    }
    */

    if ((vob->video_in_file && strlen(vob->video_in_file)>=11 && strncmp(vob->video_in_file, "/dev/video1", 11))) do_resync=0; //no resync stuff for webcams
	
    switch(vob->im_v_codec) {
      
    case CODEC_RGB:
    
      if(video_grab_init(vob->video_in_file, vob->chanid, vob->station_id, vob->im_v_width, vob->im_v_height, VIDEO_PALETTE_RGB24, verbose_flag, do_audio)<0) {
	fprintf(stderr, "error grab init\n");
	return(TC_IMPORT_ERROR);
      }

      break;

    case CODEC_YUV:

      if (vob->im_v_string && strlen (vob->im_v_string)>0) {
	  if ( (strcmp (vob->im_v_string, "yuv422")) == 0)
	      fmt = VIDEO_PALETTE_YUV422;
      }

      if(video_grab_init(vob->video_in_file, vob->chanid, vob->station_id, vob->im_v_width, vob->im_v_height, fmt, verbose_flag, do_audio)<0) {
	fprintf(stderr, "error grab init\n");
	return(TC_IMPORT_ERROR);
      }

      break;
    }
    
    vframe_pts0 =  vframe_pts = v4l_counter_init();
    if (do_audio) 
	video_drop_frames = audio_drop_frames - (int) ((vframe_pts0-aframe_pts0)*vob->fps);
    if(verbose_flag) printf("[%s] dropping %d video frames for AV sync\n ", MOD_NAME, video_drop_frames);
    
    return(0);
  }
  
  if(param->flag == TC_AUDIO) {
    
    // print out
    if(verbose_flag) printf("[%s] video4linux audio grabbing\n", MOD_NAME);
    
    //set device with vob->audio_in_file 
    if(audio_grab_init(vob->audio_in_file, vob->a_rate, vob->a_bits, vob->a_chan, verbose_flag)<0) return(TC_IMPORT_ERROR);

    aframe_pts0 = aframe_pts = v4l_counter_init();

    param->fd = NULL;
    
    return(0);
  }
  
  return(TC_IMPORT_ERROR);
}

/* ------------------------------------------------------------ 
 *
 * decode  stream
 *
 * ------------------------------------------------------------*/

static int v4l_decode(transfer_t *param, vob_t *vob)
{
  
  if(param->flag == TC_VIDEO) {

    if(!do_resync) video_drop_frames=1;

    do {
      video_grab_frame(param->buffer);
      if((verbose & TC_STATS) && vframe_cnt<MAX_DISPLAY_PTS) v4l_counter_print("VIDEO", vframe_cnt, vframe_pts0, &vframe_pts);
      
      ++vframe_cnt;
      --video_drop_frames;
      
    } while(video_drop_frames>0);

    video_drop_frames=1;
    
    return(0);
  }
  
  if(param->flag == TC_AUDIO) {

    if(!do_resync) audio_drop_frames=1;
    
    do {
      
      audio_grab_frame(param->buffer, param->size);
      if((verbose & TC_STATS) && aframe_cnt<MAX_DISPLAY_PTS) v4l_counter_print("AUDIO", aframe_cnt, aframe_pts0, &aframe_pts);

      ++aframe_cnt;
      --audio_drop_frames;

    } while(audio_drop_frames>0);
    
    audio_drop_frames=1;

    return(0);
 }
 
 return(TC_IMPORT_ERROR);

}

/* ------------------------------------------------------------ 
 *
 * close stream
 *
 * ------------------------------------------------------------*/

static int v4l_close(transfer_t *param)
{  

  if(param->flag == TC_VIDEO) {
    
    // stop grabbing
    video_grab_close(do_audio);
    
    return(0);
  }
  
  if(param->flag == TC_AUDIO) {
    
    // stop grabbing
    audio_grab_close(do_audio);
    
    return(0);
  }
  
  return(TC_IMPORT_ERROR);
}


